#include "envy.h"
#include "envy_config.h"
#include "envy_options.h"
#include "envy_osdb.h"

static void cb_window_resized(Ecore_Evas *ee);
static int  use_video_size(Evas_Object *o);
static int  bg_key_down(void *data, int ev_type, void *event_info);
static void cb_toggle_play_pause(void *data, Evas_Object *o, const char *emission, const char *source);
static void cb_progress_dragged(void *data, Evas_Object *o, const char *emission, const char *source);
static void cb_frame_decoded(void *data, Evas_Object *obj, void *event_info);
static void parse_srt(FILE *fd);
static int  sub_comp(const void *d1, const void *d2);
static int  cb_hide_info(void *data);
static int  cb_hide_mouse(void *data);
static void cb_mouse_move(void *data, Evas *e, Evas_Object *o, void *event_info);

Envy_Config *ec;

int
main(int argc, char **argv)
{
  if (argc < 2)
    {
      fprintf(stderr, "No filename in arguments\n");
      printf("Usage %s filename\n", argv[0]);
      return EXIT_FAILURE;
    }

  /* Envy */
  envy = calloc(1, sizeof(Envy));
  envy->subs_size = 300;
  envy->sub_delay = 0;
  envy->show_sub = 1;

  /* Config */
  ec = envy_config_get();

  envy->filename = envy_parse_opt(argc, argv, ec);

  /* Initialization */
  if (!ecore_init()) return -1;
  edje_init();
  ecore_app_args_set(argc, (const char **)argv);

  if (!ecore_evas_init()) return -1;
  envy->ee = ecore_evas_gl_x11_new(NULL, 0, 0, 0, 640, 480);

  if (!envy->ee)
    {
      fprintf(stderr, "AAAAAAAAAAArg ecore_evas_gl doesn't work.\n");
      return -1;
    }

  /* Window */
  ecore_evas_title_set(envy->ee, "Envy");
  ecore_evas_name_class_set(envy->ee, "envy", "envy");
  ecore_evas_show(envy->ee);
  ecore_evas_callback_resize_set(envy->ee, cb_window_resized);

  /* Evas */
  envy->evas = ecore_evas_get(envy->ee);
  evas_image_cache_set(envy->evas, 8 * 2048);
  edje_frametime_set(1.0 / 30.0);

  /* Edje */
  envy->bg = edje_object_add(envy->evas);
  if (!access(ec->theme, F_OK))
    edje_object_file_set(envy->bg, ec->theme, "player");
  else
    fprintf(stderr, "Can't find theme file\n");
  evas_object_move(envy->bg, 0, 0);
  evas_object_resize(envy->bg, 640, 480);
  evas_object_show(envy->bg);

  /* Black Background */
  envy->black_bg = evas_object_rectangle_add(envy->evas);
  evas_object_color_set(envy->black_bg, 0, 0, 0, 255);
  evas_object_show(envy->black_bg);

  /* Balloon */
  envy->balloon = evas_object_image_add(envy->evas);
  evas_object_image_file_set(envy->balloon, PACKAGE_DATA_DIR "/pixmaps/balloon.png", NULL);
  evas_object_image_fill_set(envy->balloon, 0, 0, 186, 137);
  evas_object_move(envy->balloon, 237, 182);
  evas_object_resize(envy->balloon, 186, 137);
  evas_object_show(envy->balloon);

  /* Video */
  Evas_Object *o;
  o = emotion_object_add(envy->evas);
  if (!emotion_object_init(o, "emotion_decoder_xine.so")) return EXIT_FAILURE;
  emotion_object_file_set(o, envy->filename);
  emotion_object_smooth_scale_set(o, 1);
  evas_object_move(o, 0, 0);
  evas_object_resize(o, 800, 600);
  evas_object_show(o);
  envy->video = o;

  /* Resize the window at the video size */
  ecore_timer_add(0.5, (int (*)(void *))use_video_size, o);

  /* Callbacks */
  edje_object_signal_callback_add(envy->bg, "play_pause_toggled", "*", cb_toggle_play_pause, o);
  edje_object_signal_callback_add(envy->bg, "drag", "progress", cb_progress_dragged, o);
  ecore_event_handler_add(ECORE_X_EVENT_KEY_DOWN, bg_key_down, o);
  evas_object_smart_callback_add(o, "frame_decode", cb_frame_decoded, o);
  evas_object_event_callback_add(o, EVAS_CALLBACK_MOUSE_MOVE, cb_mouse_move, envy->ee);
  evas_object_event_callback_add(envy->black_bg, EVAS_CALLBACK_MOUSE_MOVE, cb_mouse_move, envy->ee);
  
  edje_object_signal_emit(envy->bg, "play", "player");

  /* Information */
  envy->info = evas_object_text_add(envy->evas);
  evas_object_color_set(envy->info, 255, 255, 255, 255);
  evas_object_text_font_set(envy->info, "vera", 17);
  evas_object_show(envy->info);

  /* Sub */
  envy->sub = evas_object_textblock_add(envy->evas);
  evas_object_color_set(envy->sub, 255, 255, 255, 255);
  evas_object_move(envy->sub, 100, 400);
  evas_object_resize(envy->sub, 800, 200);
  evas_object_pass_events_set(envy->sub, 1);
  evas_object_show(envy->sub);

  /* Textblock Style */
  Evas_Textblock_Style *style;
  char st[512];
  snprintf(st, sizeof(st),
           "DEFAULT='font=vera font_size=20 align=center valign=bottom "
           "color=#ffffff wrap=word style=glow glow_color=#000000'"
           "br='\n'");
  style = evas_textblock_style_new();
  evas_textblock_style_set(style, st);
  evas_object_textblock_style_set(envy->sub, style);

  evas_object_layer_set(envy->info, 5);
  evas_object_layer_set(envy->bg, 4);
  evas_object_layer_set(envy->sub, 3);
  evas_object_layer_set(envy->video, 2);
  evas_object_layer_set(envy->balloon, 6);
  evas_object_layer_set(envy->black_bg, 0);


  emotion_object_play_set(envy->video, 1);

  ecore_main_loop_begin();
  return EXIT_SUCCESS;
}

void
envy_load_subtitle(const char *subfile)
{
  FILE *fd;
  fd = fopen(subfile, "r");
  if (fd == NULL)
    {
      perror("Reading subfile");
      return;
    }
  parse_srt(fd);
  fclose(fd);
}

void
envy_osdb_process(void *data)
{
  Envy_Osdb *eo = data;

  evas_object_text_text_set(envy->info, "Login in OSDB..."); 
  envy_osdb_login(eo, "en");
  evas_object_text_text_set(envy->info, "Searching...");
  if (envy_osdb_search(eo, eo->lang, eo->moviehash, eo->moviesize))
    {
      evas_object_text_text_set(envy->info, "Downloading...");
      if (envy_osdb_download(eo, eo->IDSubtitleFile))
        envy_load_subtitle(eo->subfilename);
    }
  evas_object_text_text_set(envy->info, "Logout of OSDB...");
  envy_osdb_logout(eo);
  evas_object_text_text_set(envy->info, "");
  envy_osdb_free(eo);


  return;
}

static void
cb_window_resized(Ecore_Evas *ee)
{
  Evas_Coord w, h;

  evas_output_viewport_get(envy->evas, NULL, NULL, &w, &h);
  evas_object_resize(envy->bg, w, h);
  evas_object_resize(envy->black_bg, w, h);
  if (!envy->fullscreen)
    {
      evas_object_move(envy->video, 0, 0);
      evas_object_resize(envy->video, w, h);
    }

  /* Set the correct size for the subtites */

  Evas_Textblock_Style *style;
  char st[512];
  snprintf(st, sizeof(st),
           "DEFAULT='font=vera font_size=%d align=center valign=bottom "
           "color=#ffffff wrap=word style=glow glow_color=#000000'"
           "br='\n'", h/20);
  evas_textblock_style_free((Evas_Textblock_Style *)evas_object_textblock_style_get(envy->sub));
  style = evas_textblock_style_new();
  evas_textblock_style_set(style, st);
  evas_object_textblock_style_set(envy->sub, style);
}

static int
use_video_size(Evas_Object *o)
{
  if (!emotion_object_video_handled_get(o))
    return 1;

  /* Set video size */
  int w, h;
  emotion_object_size_get(o, &w, &h);
  if (w <= 1 && h <= 1) return 1;

  ecore_evas_resize(envy->ee, w, h);
  evas_object_resize(envy->bg, w, h);

  evas_object_hide(envy->balloon);

  return 0;
}

static int
bg_key_down(void *data, int ev_type, void *event_info)
{
  Evas_Object *o = data;
  Ecore_X_Event_Key_Down *ev;

  ev = event_info;
  /* Fullscreen */
  if (!strcmp(ev->keyname, "f"))
    {
      if (!ecore_evas_fullscreen_get(envy->ee))
        {
          envy->fullscreen = 1;
          ecore_evas_fullscreen_set(envy->ee, 1);

 //         edje_object_part_unswallow(envy->bg, o);
 //         edje_object_part_swallow(envy->bg, "video", envy->black_bg);

          double ratio = 1;
          Ecore_X_Screen_Size size;
          size = ecore_x_randr_current_screen_size_get(ecore_x_window_root_first_get());

          evas_object_move(envy->black_bg, 0, 0);
          evas_object_resize(envy->black_bg, size.width, size.height);

          ratio = emotion_object_ratio_get(o);
          evas_object_resize(o, size.width, size.width/ratio);
          evas_object_move(o, 0, (size.height - size.width/ratio)/2);
//          evas_object_show(o);
          ecore_evas_ignore_events_set(envy->ee, 0);
        }
      else
        {
          envy->fullscreen = 0;
          ecore_evas_fullscreen_set(envy->ee, 0);
 //         edje_object_part_unswallow(envy->bg, black_bg);
 //         edje_object_part_swallow(envy->bg, "video", o);
 //         evas_object_stack_above(envy->sub, envy->bg);
 //         evas_object_hide(envy->black_bg);
        }
    }
  /* Toggle play/pause */
  else if (!strcmp(ev->keyname, "space"))
    {
      if (emotion_object_play_get(data))
        {
          emotion_object_play_set(data, 0);
          edje_object_signal_emit(envy->bg, "pause", "player");
        }
      else
        {
          emotion_object_play_set(data, 1);
          edje_object_signal_emit(envy->bg, "play", "player");
        }
    }
  else if (!strcmp(ev->keyname, "Left"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos-10);
    }
  else if (!strcmp(ev->keyname, "Right"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos+10);
    }
  else if (!strcmp(ev->keyname, "Down"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos-60);
    }
  else if (!strcmp(ev->keyname, "Up"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos+60);
    }
  else if (!strcmp(ev->keyname, "Next"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos-600);
    }
  else if (!strcmp(ev->keyname, "Prior"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos+600);
    }
  else if (!strcmp(ev->keyname, "q"))
    {
      exit(EXIT_SUCCESS);
    }
  else if (!strcmp(ev->keyname, "x"))
    {
      char info[512];
      envy->sub_delay += 100./1000.;
      snprintf(info, sizeof(info),  "Sub: %.0f ms", envy->sub_delay*1000);
      evas_object_text_text_set(envy->info, info);
      evas_object_show(envy->info);
      if (envy->info_timer) ecore_timer_del(envy->info_timer);
      envy->info_timer = ecore_timer_add(1.0, cb_hide_info, NULL);
    }
  else if (!strcmp(ev->keyname, "y"))
    {
      char info[512];
      envy->sub_delay -= 100./1000.;
      snprintf(info, sizeof(info),  "Sub: %.0f ms", envy->sub_delay*1000);
      evas_object_text_text_set(envy->info, info);
      evas_object_show(envy->info);
      if (envy->info_timer) ecore_timer_del(envy->info_timer);
      envy->info_timer = ecore_timer_add(1.0, cb_hide_info, NULL);
    }
  else if (!strcmp(ev->keyname, "s"))
    {
      char info[512];
      if (!envy->subs) return 0;
      envy->sub_delay = envy->subs[0].start - emotion_object_position_get(o);
      snprintf(info, sizeof(info),  "Sub: %.0f ms", envy->sub_delay*1000);
      evas_object_text_text_set(envy->info, info);
      evas_object_show(envy->info);
      if (envy->info_timer) ecore_timer_del(envy->info_timer);
      envy->info_timer = ecore_timer_add(1.0, cb_hide_info, NULL);
    }
  else if (!strcmp(ev->keyname, "v"))
    {
      if (envy->show_sub)
        {
          envy->show_sub = 0;
          evas_object_text_text_set(envy->info, "Subtitles disabled");
        }
      else
        {
          envy->show_sub = 1;
          evas_object_text_text_set(envy->info, "Subtitles enabled");
        }
      evas_object_show(envy->info);
      if (envy->info_timer) ecore_timer_del(envy->info_timer);
      envy->info_timer = ecore_timer_add(1.0, cb_hide_info, NULL);
    }
  else
    {
      printf("No bindings for %s\n", ev->keyname);
    }

  return 0;
}

static void
cb_toggle_play_pause(void *data, Evas_Object *o, const char *emission, const char *source)
{
  if (emotion_object_play_get(data))
    {
      emotion_object_play_set(data, 0);
      edje_object_signal_emit(o, "pause", "player");
    }
  else
    {
      emotion_object_play_set(data, 1);
      edje_object_signal_emit(o, "play", "player");
    }
}

#define THE_GOOD_ONE(x, y) (x == 1 ? y : x)
static void
cb_progress_dragged(void *data, Evas_Object *o, const char *emission, const char *source)
{
  if (emotion_object_seekable_get(data))
    {
      double x, y, length;

      edje_object_part_drag_value_get(o, source, &x, &y);
      length = emotion_object_play_length_get(data);
      emotion_object_position_set(data, length*THE_GOOD_ONE(x, y));
    }
  else
    {
      printf("Video isn't seekable\n");
    }
}

static void
cb_frame_decoded(void *data, Evas_Object *obj, void *event_info)
{
  double pos, len;

  pos = emotion_object_position_get(obj);
  len = emotion_object_play_length_get(obj);

  edje_object_part_drag_value_set(envy->bg, "progress", pos/len, pos/len);

  if (!envy->subs) return;

  Sub_Srt key, *current_sub = NULL;
  key.start = pos + envy->sub_delay;
  evas_object_textblock_clear(envy->sub);
  if (envy->show_sub == 0) return;
  current_sub = bsearch(&key, envy->subs, envy->subs_size, sizeof(Sub_Srt), sub_comp);
  if (current_sub)
    {
      int w, h;
      int vw, vh;
      evas_object_textblock_text_markup_set(envy->sub, current_sub->txt);
//      evas_object_textblock_text_markup_set(envy->sub, ecore_strbuf_string_get(current_sub->txt));
      evas_object_textblock_size_formatted_get(envy->sub, &w, &h);
      ecore_evas_geometry_get(envy->ee, NULL, NULL, &vw, &vh);
      evas_object_resize(envy->sub, vw, h);
      evas_object_move(envy->sub, 0, vh - h - ec->sub_padding);
    } 
}

static int
sub_comp(const void *d1, const void *d2)
{
  const Sub_Srt *sub1, *sub2;
  sub1 = d1;
  sub2 = d2;
  if (sub2->start <= sub1->start && sub2->end > sub1->start)
    return 0;
  else if (sub2->start <= sub1->start && sub2->end < sub1->start)
    return 1;
  else 
    return -1;
}

static void
parse_srt(FILE *fd)
{
  char c = '\0';
  char *id_str;
  char *start_str;
  char *end_str;
  int i = 0;
  int id_size = 5;

  id_str    = calloc(id_size, sizeof(char));
  start_str = calloc(13, sizeof(char));
  end_str   = calloc(13, sizeof(char));

  if (fd == NULL) return;

  /* Get to the end */
  fseek(fd, 0, SEEK_END);
  /* Skip the last empty line */
  fseek(fd, -2L, SEEK_CUR);
  /* Get back until we find a blank line */
  while (i != 3)
    {
      fseek(fd, -1L, SEEK_CUR);
      c = getc(fd);
      if ('\r' == c || '\n' == c)
        i++;
      else
        i = 0;
      ungetc(c, fd);
    }
  fgetc(fd);
  fgetc(fd);
  /* The last id will serv as base size for the subs sarray */
  fscanf(fd, "%d\r\n", &i);
  rewind(fd);
  if (i > 0) envy->subs_size = i;
  envy->subs = malloc(envy->subs_size * sizeof(Sub_Srt));
//  printf("%d\n", envy->subs_size);
  i = 0;

  while (!feof(fd))
    {
      int ret;
      
      ret = fscanf(fd, "%d\r\n", &(envy->subs[i].id));
      if (ret == EOF) break;
      /* XXX Doesn't Work
      if (i >= envy->subs_size)
        {
          printf("coin\n");
          envy->subs_size = envy->subs_size * 2;
          envy->subs = realloc(envy->subs, envy->subs_size * sizeof(Sub_Srt));
        }
      */
//      printf("id    : %d\n", envy->subs[i].id );

      int start_hour, start_min, start_sec, start_ms;
      int end_hour, end_min, end_sec, end_ms;
      envy->subs[i].start = 0;
      ret = fscanf(fd, "%d:%d:%d,%d --> %d:%d:%d,%d\r\n", 
                   &start_hour, &start_min, &start_sec, &start_ms,
                   &end_hour, &end_min, &end_sec, &end_ms);
      if (ret == EOF) break;
      envy->subs[i].start = 0;
      envy->subs[i].start += start_hour*3600;
      envy->subs[i].start += start_min*60;
      envy->subs[i].start += start_sec;
      envy->subs[i].start += start_ms/1000.;
      envy->subs[i].end = 0;
      envy->subs[i].end += end_hour*3600;
      envy->subs[i].end += end_min*60;
      envy->subs[i].end += end_sec;
      envy->subs[i].end += end_ms/1000.;
//      printf("start : %f\n",  envy->subs[i].start);
//      printf("end   : %f\n",  envy->subs[i].end);

      char str[256];
      unsigned int txt_size = 0;
      int len = 0;
      char *txt = NULL;
      int j = 0;
      txt = calloc(256, sizeof(char));
      while ((len = fscanf(fd, "%255[^\r\n]%*c%*c", str)) != 0 && len != EOF)
        {
          if (txt_size == 0)
            {
              txt = strcpy(txt, str);
              txt_size = strlen(str);
              /*
              txt = ecore_strbuf_new();
              ecore_strbuf_append(txt, str);
              */
            }
          else
            {
              /*
              ecore_strbuf_append(txt, "<br>");
              ecore_strbuf_append(txt, str);
              */
              if (txt_size + strlen(str) + 5 > 256)
                {
                  fprintf(stderr, "What the fuck ? More than 256char on a sub !\n");
                  break;
                }
              txt = strcat(txt, "<br>");
              txt = strcat(txt, str);
              txt_size += strlen(str) + 4;
            }
        } 
      txt = realloc(txt, txt_size + 1);
      envy->subs[i].txt = txt;
//      printf("txt   : %s\n",  envy->subs[i].txt);

      i++;
    }
}

static void 
cb_mouse_move(void *data, Evas *e, Evas_Object *o, void *event_info)
{
  Ecore_Evas *ee = data;
  Ecore_X_Window win;
  win = ecore_evas_gl_x11_window_get(ee);
  ecore_x_window_cursor_show(win, 1);

  if (envy->mouse_timer) ecore_timer_del(envy->mouse_timer);
  envy->mouse_timer = ecore_timer_add(2.0, cb_hide_mouse, ee);
}

static int
cb_hide_info(void *data)
{
  evas_object_hide(envy->info);
  evas_object_text_text_set(envy->info, NULL);
  envy->info_timer = NULL;
  return 0;
}

static int
cb_hide_mouse(void *data)
{
  Ecore_X_Window win;
  win = ecore_evas_gl_x11_window_get(data);
  ecore_x_window_cursor_show(win, 0);
  envy->mouse_timer = NULL;
  return 0;
}
