#include "envy.h"
#include "envy_b64.h"
#include "envy_osdb.h"

static uint64_t _compute_hash(FILE *fd);

/******************
 *  Macro         *
 ******************/

#ifdef DEBUG
#define _dbg_printf(format...) printf(format)
#else
#define _dbg_printf(format...)
#endif

#define MAX(x, y) (((x) > (y)) ? (x) : (y))

/* Weak Assertion, return if false */
#define WASSERT(x, ret, msg...) \
  do \
    { \
      int errnum = errno; \
      if (!(x)) \
        { \
          error(0, errnum, msg); \
          return ret; \
        } \
    } \
  while (0);

/* If a fault occured in the rpc request, call func and return with ret value */
#define RETURN_IF_RPCFAULT(func, ret) \
  do \
    { \
        if (eo->env.fault_occurred) \
          { \
            fprintf(stderr, "XMLRPC failed. %s (XML-RPC fault code %d)\n", \
                    eo->env.fault_string, eo->env.fault_code); \
            func; \
            return ret; \
          } \
    } \
 while (0);

#define ENVY_OSDB_RPC_GET(structP, type, key) \
  xmlrpc_struct_find_value(&(eo->env), structP, #key, &valueP); \
  RETURN_IF_RPCFAULT(,0); \
  xmlrpc_read_##type(&(eo->env), valueP, &(eo->key)); \
  RETURN_IF_RPCFAULT(,0); 

#define LOCAL_VAR_RPC_GET(structP, type, key) \
  xmlrpc_struct_find_value(&(eo->env), structP, #key, &valueP); \
  RETURN_IF_RPCFAULT(,0); \
  xmlrpc_read_##type(&(eo->env), valueP, &key); \
  RETURN_IF_RPCFAULT(,0); 

#define HANDLE_ASSERT(x) \
  if (!(x)) \
    { \
      _dbg_printf("RPC with method '%s' at URL '%s' " \
                  "has failed\n", method_name, eo->server_url); \
      return 0; \
    }

#define SUBINFO_RPC_GET(key) \
  const char *key;\
  xmlrpc_struct_find_value(&(eo->env), subinfo, #key, &valueP); \
  RETURN_IF_RPCFAULT(,0); \
  xmlrpc_read_string(&(eo->env), valueP, &key); \
  RETURN_IF_RPCFAULT(,0); \
  _dbg_printf("%s: %s\n", #key, key); 

/*****************************
 *  Public Functions         *
 *****************************/
unsigned int
envy_moviesize_get(Envy_Osdb *eo, const char *filename)
{
  struct stat st;
  WASSERT(stat(filename, &st) == 0, 0, "Stating %s", filename);
  if (eo != NULL) eo->moviesize = st.st_size;
  return st.st_size;
}

/* Return the movie hash of filename */
uint64_t
envy_moviehash_get(Envy_Osdb *eo, const char *filename)
{
  FILE *fd;
  uint64_t moviehash;

  fd = fopen(filename, "rb");
  WASSERT(fd != NULL, 0, "Openning %s", filename);
  moviehash = _compute_hash(fd);
  _dbg_printf("Movie Hash is : %"PRIx64"\n", moviehash);
  fclose(fd);
  if (eo != NULL) eo->moviehash = moviehash;

  return moviehash;
}

Envy_Osdb *
envy_osdb_new(const char *server_url)
{
  Envy_Osdb *eo;

  eo = calloc(1, sizeof(Envy_Osdb));

  xmlrpc_env_init(&(eo->env));
  RETURN_IF_RPCFAULT(free(eo), NULL);
  xmlrpc_client_init2(&(eo->env), XMLRPC_CLIENT_NO_FLAGS, PACKAGE, VERSION, NULL, 0);
  RETURN_IF_RPCFAULT(free(eo), NULL);

  eo->server_url = strdup(server_url);
  strncpy(eo->lang, "eng", 3);
  eo->lang[3] = '\0';
  eo->keep_osdb_subfilename = 0;

  return eo;
}

int
envy_osdb_login(Envy_Osdb *eo, const char *lang2c)
{
  xmlrpc_value *resultP;
  xmlrpc_value *valueP;
  const char *status;
  double seconds;
  const char *method_name = "LogIn";

  resultP = xmlrpc_client_call(&(eo->env), eo->server_url, method_name, 
                               "(ssss)", "", "", lang2c, PACKAGE);
  RETURN_IF_RPCFAULT(, 0);
  _dbg_printf("Making XMLRPC call to server url '%s' method '%s' "
              "...\n", eo->server_url, "LogIn");

  HANDLE_ASSERT(resultP);

  LOCAL_VAR_RPC_GET(resultP, string, status)
  LOCAL_VAR_RPC_GET(resultP, double, seconds)
  _dbg_printf("%s: %s\n", "status", status); 
  _dbg_printf("%s: %f\n", "seconds", seconds); 

  HANDLE_ASSERT(strncmp(status, "200", 3) == 0);

  ENVY_OSDB_RPC_GET(resultP, string, token)
  _dbg_printf("%s: %s\n", "token", eo->token); 

  _dbg_printf("RPC with method '%s' at URL '%s' "
              "has succefully completed\n", method_name, eo->server_url);
  return 1;
}

int
envy_osdb_search(Envy_Osdb *eo, const char *lang3c, uint64_t moviehash, unsigned int moviesize)
{
  xmlrpc_value *resultP;
  char moviehash_str[17];
  xmlrpc_value *valueP;
  xmlrpc_value *result_array;
  xmlrpc_value *subinfo;
  double seconds;
  const char *method_name = "SearchSubtitles";

  sprintf(moviehash_str, "%"PRIx64, moviehash);
  resultP = xmlrpc_client_call(&(eo->env), eo->server_url, method_name, 
                               "(s({s:s,s:s,s:i}))", eo->token, 
                               "sublanguageid", lang3c, 
                               "moviehash", moviehash_str, 
                               "moviebytesize", moviesize);
  RETURN_IF_RPCFAULT(, 0);
  _dbg_printf("Making XMLRPC call to server url '%s' method '%s' "
              "...\n", eo->server_url, "SearchSubtitles");

  HANDLE_ASSERT(resultP);

  LOCAL_VAR_RPC_GET(resultP, double, seconds)
  _dbg_printf("%s: %f\n", "seconds", seconds); 
  xmlrpc_struct_find_value(&(eo->env), resultP, "data", &result_array); 
  RETURN_IF_RPCFAULT(,0);
  HANDLE_ASSERT(xmlrpc_value_type(result_array) == XMLRPC_TYPE_ARRAY);
  xmlrpc_array_read_item(&(eo->env), result_array, 0, &subinfo); 
  RETURN_IF_RPCFAULT(,0);

/* Not sent by the server 
   SUBINFO_RPC_GET(MovieFrames);
 */
  SUBINFO_RPC_GET(IDSubMovieFile);
  SUBINFO_RPC_GET(MovieHash);
  SUBINFO_RPC_GET(MovieByteSize);
  SUBINFO_RPC_GET(MovieTimeMS);
  SUBINFO_RPC_GET(SubFileName);
  SUBINFO_RPC_GET(SubActualCD);
  SUBINFO_RPC_GET(SubSize);
  SUBINFO_RPC_GET(SubHash);
  SUBINFO_RPC_GET(IDSubtitle);
  SUBINFO_RPC_GET(IDSubtitleFile);
  SUBINFO_RPC_GET(UserID);
  SUBINFO_RPC_GET(SubLanguageID);
  SUBINFO_RPC_GET(SubFormat);
  SUBINFO_RPC_GET(SubSumCD);
  SUBINFO_RPC_GET(SubAuthorComment);
  SUBINFO_RPC_GET(SubAddDate);
  SUBINFO_RPC_GET(SubBad);
  SUBINFO_RPC_GET(SubRating);
  SUBINFO_RPC_GET(SubDownloadsCnt);
  SUBINFO_RPC_GET(MovieReleaseName);
  SUBINFO_RPC_GET(IDMovie);
  SUBINFO_RPC_GET(IDMovieImdb);
  SUBINFO_RPC_GET(MovieName);
  SUBINFO_RPC_GET(MovieNameEng);
  SUBINFO_RPC_GET(MovieYear);
  SUBINFO_RPC_GET(MovieImdbRating);
  SUBINFO_RPC_GET(UserNickName);
  SUBINFO_RPC_GET(ISO639);
  SUBINFO_RPC_GET(LanguageName);
  SUBINFO_RPC_GET(SubDownloadLink);

  ENVY_OSDB_RPC_GET(subinfo, string, SubFileName);
  ENVY_OSDB_RPC_GET(subinfo, string, IDSubtitleFile);

  _dbg_printf("RPC with method '%s' at URL '%s' "
              "has succefully completed\n", method_name, eo->server_url);
  return 1;
}

int
envy_osdb_download(Envy_Osdb *eo, const char *id_subtitle_file)
{
  xmlrpc_value *resultP;
  xmlrpc_value *result_array;
  xmlrpc_value *sub_struct;
  xmlrpc_value *valueP;
  const char *status;
  double seconds;
  const char *method_name = "DownloadSubtitles";

  HANDLE_ASSERT(eo->IDSubtitleFile != NULL);

  _dbg_printf("Making XMLRPC call to server url '%s' method '%s' "
              "...\n", eo->server_url, "DownloadSubtitles");
  resultP = xmlrpc_client_call(&(eo->env), eo->server_url, method_name, 
                               "(s(s))",  eo->token, eo->IDSubtitleFile);
  RETURN_IF_RPCFAULT(, 0);

  HANDLE_ASSERT(resultP);

  LOCAL_VAR_RPC_GET(resultP, string, status)
  LOCAL_VAR_RPC_GET(resultP, double, seconds)
  _dbg_printf("%s: %s\n", "status", status); 
  _dbg_printf("%s: %f\n", "seconds", seconds); 

  HANDLE_ASSERT(strncmp(status, "200", 3) == 0);

  xmlrpc_struct_find_value(&(eo->env), resultP, "data", &result_array); 
  RETURN_IF_RPCFAULT(,0);

  HANDLE_ASSERT(xmlrpc_value_type(result_array) == XMLRPC_TYPE_ARRAY);
  xmlrpc_array_read_item(&(eo->env), result_array, 0, &sub_struct); 
  RETURN_IF_RPCFAULT(,0);

  xmlrpc_struct_find_value(&(eo->env), sub_struct, "data", &valueP); 
  RETURN_IF_RPCFAULT(,0);

  /* We get the b64 file */
  const char *sub_base64;
  xmlrpc_read_string(&(eo->env), valueP, &sub_base64);

  /* Decoding it and getting the gzip file */
  size_t cb;
  void *sub_gziped;
  cb = b64_decode(sub_base64, strlen(sub_base64), NULL, 0);
  sub_gziped = malloc(cb);
  b64_decode(sub_base64, strlen(sub_base64), sub_gziped, cb);
 
  /* We write the gzip file on the hdd */
  FILE *dest;
  char gzipfile[PATH_MAX];
  strcpy(gzipfile, "/tmp/");
  strcat(gzipfile, eo->SubFileName);
  strcat(gzipfile, ".gz");
  dest = fopen(gzipfile, "wb");
  WASSERT(dest != NULL, 0, "Error opening the gzip subfile for writting\n");
  if (fwrite(sub_gziped, 1, cb, dest) != cb)
    {
      fprintf(stderr, "Error writting the gziped subfile\n");
      fclose(dest);
      return 0;
    }
  fclose(dest);

  /* Decompress the subfile */
  char sub[4096];
  char subfile[PATH_MAX];
  size_t ret;
  if (!eo->subfilename || eo->keep_osdb_subfilename)
    {
      eo->subfilename = calloc(strlen(envy->dirname) + strlen(eo->SubFileName) + 2, 1);
      strcpy(eo->subfilename, envy->dirname);
      strcat(eo->subfilename, "/");
      strcat(eo->subfilename, eo->SubFileName);
    }
  gzFile src = gzopen(gzipfile, "rb");
  WASSERT(src != NULL, 0, "Error opening the gzip subfile for reading\n");
  dest = fopen(eo->subfilename, "wb");
  WASSERT(src != NULL, 0, "Error opening the subfile for writting\n");
  while ((ret = gzread(src, sub, 4096)) != 0)
    {
      if (fwrite(sub, 1, ret, dest) != ret)
        {
          fprintf(stderr, "Error writting the subfile\n");
          gzclose(src);
          fclose(dest);
          return 0;
        }
    }
  gzclose(src);
  fclose(dest);
  WASSERT(unlink(gzipfile) == 0, 0, "Error while deleting the gziped subfile\n");

  printf("RPC with method '%s' at URL '%s' "
         "has completed\n", method_name, eo->server_url);
  return 1;
}

int
envy_osdb_logout(Envy_Osdb *eo)
{
  xmlrpc_value *resultP;
  xmlrpc_value *valueP;
  const char *status;
  double seconds;
  const char *method_name = "LogOut";

  resultP = xmlrpc_client_call(&(eo->env), eo->server_url, method_name, "(s)", eo->token);
  RETURN_IF_RPCFAULT(, 0);
  _dbg_printf("Making XMLRPC call to server url '%s' method '%s' "
              "...\n", eo->server_url, "LogOut");

  HANDLE_ASSERT(resultP);

  LOCAL_VAR_RPC_GET(resultP, string, status)
  LOCAL_VAR_RPC_GET(resultP, double, seconds)
  _dbg_printf("%s: %s\n", "status", status); 
  _dbg_printf("%s: %f\n", "seconds", seconds); 

  HANDLE_ASSERT(strncmp(status, "200", 3) == 0);

  _dbg_printf("RPC with method '%s' at URL '%s' "
              "has succefully completed\n", method_name, eo->server_url);
  return 1;
}

void
envy_osdb_free(Envy_Osdb *eo)
{
  xmlrpc_env_clean(&(eo->env));
  xmlrpc_client_cleanup();
  free(eo->server_url);
  free(eo);
}

/*****************************
 *  Private Functions        *
 *****************************/
static uint64_t 
_compute_hash(FILE *fd)
{
  uint64_t hash, fsize, tmp, i;

  fseek(fd, 0, SEEK_END);
  fsize = ftell(fd);
  rewind(fd);

  hash = fsize;
  for(tmp = 0, i = 0; i < 65536/sizeof(tmp) && fread(&tmp, sizeof(tmp), 1, fd); hash += tmp, i++);
  fseek(fd, MAX(0, fsize - 65536), SEEK_SET);
  for(tmp = 0, i = 0; i < 65536/sizeof(tmp) && fread(&tmp, sizeof(tmp), 1, fd); hash += tmp, i++);

  return hash;
}
